import base64
import unittest
from pathlib import Path
from pprint import pprint
from time import sleep

from oldaplib.src.objectfactory import ResourceInstanceFactory
from oldaplib.src.connection import Connection
from oldaplib.src.enums.action import Action
from oldaplib.src.enums.datapermissions import DataPermission
from oldaplib.src.enums.language import Language
from oldaplib.src.helpers.attributechange import AttributeChange
from oldaplib.src.helpers.context import Context
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.helpers.oldaperror import OldapErrorNotFound, OldapErrorValue, OldapErrorNoPermission, OldapErrorInUse
from oldaplib.src.permissionset import PermissionSet
from oldaplib.src.project import Project
from oldaplib.src.user import User
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_anyuri import Xsd_anyURI
from oldaplib.src.xsd.xsd_base64binary import Xsd_base64Binary
from oldaplib.src.xsd.xsd_boolean import Xsd_boolean
from oldaplib.src.xsd.xsd_byte import Xsd_byte
from oldaplib.src.xsd.xsd_date import Xsd_date
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.xsd.xsd_datetimestamp import Xsd_dateTimeStamp
from oldaplib.src.xsd.xsd_decimal import Xsd_decimal
from oldaplib.src.xsd.xsd_double import Xsd_double
from oldaplib.src.xsd.xsd_duration import Xsd_duration
from oldaplib.src.xsd.xsd_float import Xsd_float
from oldaplib.src.xsd.xsd_gday import Xsd_gDay
from oldaplib.src.xsd.xsd_gmonth import Xsd_gMonth
from oldaplib.src.xsd.xsd_gmonthday import Xsd_gMonthDay
from oldaplib.src.xsd.xsd_gyear import Xsd_gYear
from oldaplib.src.xsd.xsd_gyearmonth import Xsd_gYearMonth
from oldaplib.src.xsd.xsd_hexbinary import Xsd_hexBinary
from oldaplib.src.xsd.xsd_id import Xsd_ID
from oldaplib.src.xsd.xsd_idref import Xsd_IDREF
from oldaplib.src.xsd.xsd_int import Xsd_int
from oldaplib.src.xsd.xsd_integer import Xsd_integer
from oldaplib.src.xsd.xsd_language import Xsd_language
from oldaplib.src.xsd.xsd_long import Xsd_long
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_negativeinteger import Xsd_negativeInteger
from oldaplib.src.xsd.xsd_nmtoken import Xsd_NMTOKEN
from oldaplib.src.xsd.xsd_nonnegativeinteger import Xsd_nonNegativeInteger
from oldaplib.src.xsd.xsd_nonpositiveinteger import Xsd_nonPositiveInteger
from oldaplib.src.xsd.xsd_normalizedstring import Xsd_normalizedString
from oldaplib.src.xsd.xsd_positiveinteger import Xsd_positiveInteger
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_short import Xsd_short
from oldaplib.src.xsd.xsd_string import Xsd_string
from oldaplib.src.xsd.xsd_time import Xsd_time
from oldaplib.src.xsd.xsd_token import Xsd_token
from oldaplib.src.xsd.xsd_unsignedbyte import Xsd_unsignedByte
from oldaplib.src.xsd.xsd_unsignedint import Xsd_unsignedInt
from oldaplib.src.xsd.xsd_unsignedlong import Xsd_unsignedLong
from oldaplib.src.xsd.xsd_unsignedshort import Xsd_unsignedShort


def find_project_root(current_path):
    # Climb up the directory hierarchy and check for a marker file
    path = Path(current_path).absolute()
    while not (path / 'pyproject.toml').exists():
        if path.parent == path:
            # Root of the filesystem, file not found
            raise RuntimeError('Project root not found')
        path = path.parent
    return path

class TestObjectFactory(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        project_root = find_project_root(__file__)
        cls._context = Context(name="DEFAULT")

        cls._connection = Connection(server='http://localhost:7200',
                                     repo="oldap",
                                     userId="rosenth",
                                     credentials="RioGrande",
                                     context_name="DEFAULT")
        cls._unpriv = Connection(server='http://localhost:7200',
                                 repo="oldap",
                                 userId="unknown",
                                 credentials="RioGrande",
                                 context_name="DEFAULT")
        cls._context['test'] = 'http://oldap.org/test#'
        cls._context.use('test')

        cls._connection.clear_graph(Xsd_QName('test:shacl'))
        cls._connection.clear_graph(Xsd_QName('test:onto'))
        cls._connection.clear_graph(Xsd_QName('test:data'))
        cls._connection.clear_graph(Xsd_QName('oldap:admin'))

        file = project_root / 'oldaplib' / 'ontologies' / 'admin.trig'
        cls._connection.upload_turtle(file)
        sleep(1)  # upload may take a while...

        file = project_root / 'oldaplib' / 'testdata' / 'objectfactory_test.trig'
        cls._connection.upload_turtle(file)

        sleep(1)  # upload may take a while...

        user = User.read(cls._connection, "rosenth")
        user.hasPermissions.add(Iri('oldap:GenericUpdate'))
        user.update()

        ps = PermissionSet(con=cls._connection,
                           permissionSetId="testNoUpdate",
                           label=LangString("testNoUpdate@en"),
                           comment=LangString("Testing PermissionSet@en"),
                           givesPermission=DataPermission.DATA_VIEW,
                           definedByProject="test")
        ps.create()
        cls._tps = ps.read(cls._connection, "testNoUpdate", "test")

        user = User(con=cls._connection,
                    userId=Xsd_NCName("factorytestuser"),
                    familyName="FactoryTest",
                    givenName="FactoryTest",
                    credentials="Waseliwas",
                    inProject={'oldap:Test': {}},
                    hasPermissions={ps.iri.as_qname},
                    isActive=True)
        user.create()
        cls._tuser = User.read(cls._connection, "factorytestuser")


    @classmethod
    def tearDownClass(cls):
        #cls._connection.clear_graph(Xsd_QName('oldaplib:admin'))
        #cls._connection.upload_turtle("oldaplib/ontologies/admin.trig")
        #sleep(1)  # upload may take a while...
        pass

    def test_constructor_A(self):
        project = Project.read(con=self._connection, projectIri_SName='test')
        factory = ResourceInstanceFactory(con=self._connection, project=project)
        Book = factory.createObjectInstance('Book')
        b = Book(title="Hitchhiker's Guide to the Galaxy",
                 author=Iri('test:DouglasAdams', validate=False),
                 pubDate="1995-09-27",
                 grantsPermission=Iri('oldap:GenericView'))
        self.assertEqual(b.title, {Xsd_string("Hitchhiker's Guide to the Galaxy")})
        self.assertEqual(b.author, {Iri('test:DouglasAdams')})
        self.assertEqual(b.pubDate, {Xsd_date("1995-09-27")})
        self.assertIsNotNone(b.creationDate)
        self.assertEqual(b.createdBy, Iri('https://orcid.org/0000-0003-1681-4036', validate=False))
        self.assertIsNotNone(b.lastModificationDate)
        self.assertEqual(b.lastModifiedBy, Iri('https://orcid.org/0000-0003-1681-4036', validate=False))
        b.create()
        b2 = Book.read(con=self._connection,
                       project='test',
                       iri=b.iri)
        self.assertEqual(b2.title, b.title)
        self.assertEqual(b2.author, b.author)
        self.assertEqual(b2.pubDate, b.pubDate)
        self.assertEqual(b2.createdBy, b.createdBy)
        self.assertEqual(b2.creationDate, b.creationDate)
        self.assertEqual(b2.lastModifiedBy, b.lastModifiedBy)
        self.assertEqual(b2.lastModificationDate, b.lastModificationDate)

        #
        # test unprivileged access. Should return "not found"-Error!
        #
        with self.assertRaises(OldapErrorNotFound):
            b3 = Book.read(con=self._unpriv, project='test', iri=b.iri)

        Page = factory.createObjectInstance('Page')
        p1 = Page(pageDesignation="Cover",
                  pageNum=1,
                  pageDescription=LangString("Cover page of book@en", "Vorderseite des Bucheinschlags@de"),
                  pageInBook="test:Hitchhiker",
                  grantsPermission=Iri('oldap:GenericView'))
        self.assertEqual(p1.pageDesignation, {"Cover"})
        self.assertEqual(p1.pageNum, {1})
        self.assertEqual(p1.pageDescription, LangString("Cover page of book@en", "Vorderseite des Bucheinschlags@de"))
        self.assertEqual(p1.pageInBook, {"test:Hitchhiker"})
        p1.create()
        p2 = Page.read(con=self._connection, project='test', iri=p1.iri)
        self.assertEqual(p2.pageDesignation, p1.pageDesignation)
        self.assertEqual(p2.pageNum, p1.pageNum)
        self.assertEqual(p2.pageInBook, p1.pageInBook)
        self.assertEqual(p2.createdBy, p1.createdBy)
        self.assertEqual(p2.creationDate, p1.creationDate)
        self.assertEqual(p2.lastModifiedBy, p1.lastModifiedBy)
        self.assertEqual(p2.lastModificationDate, p1.lastModificationDate)

    def test_constructor_B(self):
        factory = ResourceInstanceFactory(con=self._connection, project='test')
        AllTypes = factory.createObjectInstance('AllTypes')
        at = AllTypes(stringProp=Xsd_string("A String Prop"),
                      langStringProp=LangString("A LangString@en", "Ein Sprachtext@de"),
                      booleanProp=Xsd_boolean(1),
                      decimalProp=Xsd_decimal(1.5),
                      floatProp= Xsd_float(1.5),
                      doubleProp=Xsd_double(1.5),
                      durationProp=Xsd_duration('PT2M10S'),
                      dateTimeProp=Xsd_dateTime('2001-10-26T21:32:52'),
                      dateTimeStampProp=Xsd_dateTimeStamp('2001-10-26T21:32:52Z'),
                      timeProp=Xsd_time('21:32:52+02:00'),
                      dateProp=Xsd_date(2025, 12, 31),
                      gYearMonthProp=Xsd_gYearMonth("2020-03"),
                      gYearProp=Xsd_gYear("2020"),
                      gMonthDayProp=Xsd_gMonthDay("--02-21"),
                      gDayProp=Xsd_gDay("---01"),
                      gMonthProp=Xsd_gMonth("--10"),
                      hexBinaryProp=Xsd_hexBinary("1fab17fa"),
                      base64BinaryProp=Xsd_base64Binary(base64.b64encode(b'Waseliwas soll den das sein?')),
                      anyURIProp=Xsd_anyURI('http://www.org/test'),
                      QNameProp=Xsd_QName('prefix:name'),
                      normalizedStringProp=Xsd_normalizedString("Dies ist ein string mit $onderzeichen\" und anderen Dingen"),
                      tokenProp=Xsd_token("Dies ist ein string mit $onderzeichen und anderen Dingen"),
                      languageProp=Xsd_language("de"),
                      nCNameProp=Xsd_NCName("Xsd_NCName"),
                      nMTOKENProp=Xsd_NMTOKEN(":ein.Test"),
                      iDProp=Xsd_ID("anchor"),
                      iDREFProp=Xsd_IDREF("anchor"),
                      integerProp=Xsd_integer(1),
                      nonPositiveIntegerProp=Xsd_nonPositiveInteger(-22),
                      negativeIntegerProp=Xsd_negativeInteger(-22),
                      longProp=Xsd_long(505_801),
                      intProp=Xsd_int(505_801),
                      shortProp=Xsd_short(-2024),
                      byteProp=Xsd_byte(100),
                      nonNegativeIntegerProp=Xsd_nonNegativeInteger(202_203_204),
                      unsignedLongProp=Xsd_unsignedLong(202_203_204),
                      unsignedIntProp=Xsd_unsignedInt(20200),
                      unsignedShortProp=Xsd_unsignedShort(20200),
                      unsignedByteProp=Xsd_unsignedByte(202),
                      positiveIntegerProp=Xsd_unsignedByte(202),
                      grantsPermission=Iri('oldap:GenericView'))
        self.assertEqual(at.stringProp, {Xsd_string("A String Prop")})
        self.assertEqual(at.langStringProp, LangString("A LangString@en", "Ein Sprachtext@de"))
        self.assertEqual(at.booleanProp, Xsd_boolean(1))
        self.assertEqual(at.decimalProp, {Xsd_decimal(1.5)})
        self.assertEqual(at.floatProp, {Xsd_float(1.5)})
        self.assertEqual(at.doubleProp, {Xsd_double(1.5)})
        self.assertEqual(at.durationProp, {Xsd_duration('PT2M10S')})
        self.assertEqual(at.dateTimeProp, {Xsd_dateTime('2001-10-26T21:32:52')})
        self.assertEqual(at.dateTimeStampProp, {Xsd_dateTimeStamp('2001-10-26T21:32:52Z')})
        self.assertEqual(at.timeProp, {Xsd_time('21:32:52+02:00')})
        self.assertEqual(at.dateProp, {Xsd_date(2025, 12, 31)})
        self.assertEqual(at.gYearMonthProp, {Xsd_gYearMonth("2020-03")})
        self.assertEqual(at.gYearProp, {Xsd_gYear("2020")})
        self.assertEqual(at.gMonthDayProp, {Xsd_gMonthDay("--02-21")})
        self.assertEqual(at.gDayProp, {Xsd_gDay("---01")})
        self.assertEqual(at.gMonthProp, {Xsd_gMonth("--10")})
        self.assertEqual(at.hexBinaryProp, {Xsd_hexBinary("1fab17fa")})
        self.assertEqual(at.base64BinaryProp, {Xsd_base64Binary(base64.b64encode(b'Waseliwas soll den das sein?'))})
        self.assertEqual(at.anyURIProp, {Xsd_anyURI('http://www.org/test')})
        self.assertEqual(at.QNameProp, {Xsd_QName('prefix:name')})
        self.assertEqual(at.normalizedStringProp,
                         {Xsd_normalizedString("Dies ist ein string mit $onderzeichen\" und anderen Dingen")})
        self.assertEqual(at.tokenProp, {Xsd_token("Dies ist ein string mit $onderzeichen und anderen Dingen")})
        self.assertEqual(at.languageProp, {Xsd_language("de")})
        self.assertEqual(at.nCNameProp, {Xsd_NCName("Xsd_NCName")})
        self.assertEqual(at.nMTOKENProp, {Xsd_NMTOKEN(":ein.Test")})
        self.assertEqual(at.iDProp, {Xsd_ID("anchor")})
        self.assertEqual(at.iDREFProp, {Xsd_IDREF("anchor")})
        self.assertEqual(at.integerProp, {Xsd_integer(1)})
        self.assertEqual(at.nonPositiveIntegerProp, {Xsd_nonPositiveInteger(-22)})
        self.assertEqual(at.negativeIntegerProp, {Xsd_negativeInteger(-22)})
        self.assertEqual(at.longProp, {Xsd_long(505_801)})
        self.assertEqual(at.intProp, {Xsd_int(505_801)})
        self.assertEqual(at.shortProp, {Xsd_short(-2024)})
        self.assertEqual(at.byteProp, {Xsd_byte(100)})
        self.assertEqual(at.nonNegativeIntegerProp, {Xsd_nonNegativeInteger(202_203_204)})
        self.assertEqual(at.negativeIntegerProp, {Xsd_negativeInteger(-22)})
        self.assertEqual(at.longProp, {Xsd_long(505_801)})
        self.assertEqual(at.intProp, {Xsd_int(505_801)})
        self.assertEqual(at.shortProp, {Xsd_short(-2024)})
        self.assertEqual(at.byteProp, {Xsd_byte(100)})
        self.assertEqual(at.unsignedLongProp, {Xsd_unsignedLong(202_203_204)})
        self.assertEqual(at.unsignedIntProp, {Xsd_unsignedInt(20200)})
        self.assertEqual(at.unsignedShortProp, {Xsd_unsignedShort(20200)})
        self.assertEqual(at.unsignedByteProp, {Xsd_unsignedByte(202)})
        self.assertEqual(at.positiveIntegerProp, {Xsd_positiveInteger(202)})
        at.create()
        at2 = AllTypes.read(con=self._connection, project='test', iri=at.iri)
        self.assertEqual(at.stringProp, at2.stringProp)
        self.assertEqual(at.langStringProp, at2.langStringProp)
        self.assertEqual(at.booleanProp, at2.booleanProp)
        self.assertEqual(at.decimalProp, at2.decimalProp)
        self.assertEqual(at.floatProp, at2.floatProp)
        self.assertEqual(at.doubleProp, at2.doubleProp)
        self.assertEqual(at.durationProp, at2.durationProp)
        self.assertEqual(at.dateTimeProp, at2.dateTimeProp)
        self.assertEqual(at.dateTimeStampProp, at2.dateTimeStampProp)
        self.assertEqual(at.timeProp, at2.timeProp)
        self.assertEqual(at.dateProp, at2.dateProp)
        self.assertEqual(at.gYearMonthProp, at2.gYearMonthProp)
        self.assertEqual(at.gYearProp, at2.gYearProp)
        self.assertEqual(at.gMonthDayProp, at2.gMonthDayProp)
        self.assertEqual(at.gDayProp, at2.gDayProp)
        self.assertEqual(at.gMonthProp, at2.gMonthProp)
        self.assertEqual(at.hexBinaryProp, at2.hexBinaryProp)
        self.assertEqual(at.base64BinaryProp, at2.base64BinaryProp)
        self.assertEqual(at.anyURIProp, at2.anyURIProp)
        self.assertEqual(at.QNameProp, at2.QNameProp)
        self.assertEqual(at.normalizedStringProp, at2.normalizedStringProp)
        self.assertEqual(at.languageProp, at2.languageProp)
        self.assertEqual(at.nCNameProp, at2.nCNameProp)
        self.assertEqual(at.nMTOKENProp, at2.nMTOKENProp)
        self.assertEqual(at.iDProp, at2.iDProp)
        self.assertEqual(at.iDREFProp, at2.iDREFProp)
        self.assertEqual(at.integerProp, at2.integerProp)
        self.assertEqual(at.nonPositiveIntegerProp, at2.nonPositiveIntegerProp)
        self.assertEqual(at.negativeIntegerProp, at2.negativeIntegerProp)
        self.assertEqual(at.longProp, at2.longProp)
        self.assertEqual(at.intProp, at2.intProp)
        self.assertEqual(at.shortProp, at2.shortProp)
        self.assertEqual(at.byteProp, at2.byteProp)
        self.assertEqual(at.nonNegativeIntegerProp, at2.nonNegativeIntegerProp)
        self.assertEqual(at.negativeIntegerProp, at2.negativeIntegerProp)
        self.assertEqual(at.longProp, at2.longProp)
        self.assertEqual(at.intProp, at2.intProp)
        self.assertEqual(at.shortProp, at2.shortProp)
        self.assertEqual(at.byteProp, at2.byteProp)
        self.assertEqual(at.unsignedLongProp, at2.unsignedLongProp)
        self.assertEqual(at.unsignedIntProp, at2.unsignedIntProp)
        self.assertEqual(at.unsignedShortProp, at2.unsignedShortProp)
        self.assertEqual(at.unsignedByteProp, at2.unsignedByteProp)
        self.assertEqual(at.positiveIntegerProp, at2.positiveIntegerProp)

    def test_value_setter(self):
        factory = ResourceInstanceFactory(con=self._connection, project='test')
        SetterTester = factory.createObjectInstance('SetterTester')
        obj1 = SetterTester(stringSetter="This is a test string",
                            langStringSetter=LangString("This is a test string@de"),
                            decimalSetter=Xsd_decimal(3.14),
                            integerSetter={20200, 30300},
                            grantsPermission={Iri('oldap:GenericView'), Iri('oldap:GenericUpdate')})
        obj1.create()
        obj2 = SetterTester.read(con=self._connection, project='test', iri=obj1.iri)
        self.assertEqual(obj1.stringSetter, obj2.stringSetter)
        self.assertEqual(obj1.langStringSetter, obj2.langStringSetter)
        self.assertIsNone(obj2.booleanSetter)
        self.assertEqual(obj1.decimalSetter, obj2.decimalSetter)
        self.assertEqual(obj1.integerSetter, obj2.integerSetter)

        #obj3 = SetterTester.read(con=self._connection, project='test', iri=obj1.iri)
        with self.assertRaises(OldapErrorValue):
            obj2.stringSetter = {"This is not a statement!", "One value too much"}
        with self.assertRaises(OldapErrorValue):
            obj2.langStringSetter = LangString("In Deutsch@de", "In Italiano@it")
        with self.assertRaises(OldapErrorValue):
            obj2.decimalSetter = {Xsd_decimal(3.14159), Xsd_decimal(2.71828), Xsd_decimal(1.61803), Xsd_decimal(1.0)}
        with self.assertRaises(OldapErrorValue):
            obj2.decimalSetter = None

        obj2.stringSetter = "This is not a statement!"
        obj2.langStringSetter = LangString("In Deutsch@de", "En Français@fr")
        obj2.booleanSetter = True
        obj2.decimalSetter = {Xsd_decimal(3.14159), Xsd_decimal(2.71828), Xsd_decimal(1.61803)}
        obj2.integerSetter = None
        self.assertEqual(obj2.stringSetter, {"This is not a statement!"})
        self.assertEqual(obj2.langStringSetter, LangString("In Deutsch@de", "En Français@fr"))
        self.assertTrue(obj2.booleanSetter)
        self.assertEqual(obj2.decimalSetter, {Xsd_decimal(3.14159), Xsd_decimal(2.71828), Xsd_decimal(1.61803)})
        self.assertFalse(obj2.integerSetter)

        expected_cs = {
            Iri("test:stringSetter"): AttributeChange(old_value={Xsd_string("This is a test string")}, action=Action.REPLACE),
            Iri("test:decimalSetter"): AttributeChange(old_value={Xsd_decimal(3.14)}, action=Action.REPLACE),
            Iri("test:langStringSetter"): AttributeChange(old_value=LangString("This is a test string@de"), action=Action.REPLACE),
            Iri("test:booleanSetter"): AttributeChange(old_value=None, action=Action.CREATE),
            Iri("test:integerSetter"): AttributeChange(old_value={20200, 30300}, action=Action.DELETE)}
        self.assertEqual(obj2.changeset, expected_cs)

        obj2.update()
        obj2 = SetterTester.read(con=self._connection, project='test', iri=obj1.iri)
        self.assertEqual(obj2.stringSetter, {"This is not a statement!"})
        self.assertEqual(obj2.langStringSetter, LangString("In Deutsch@de", "En Français@fr"))
        self.assertTrue(obj2.booleanSetter)
        self.assertEqual(obj2.decimalSetter, {Xsd_decimal(3.14159), Xsd_decimal(2.71828), Xsd_decimal(1.61803)})
        self.assertFalse(obj2.integerSetter)

    def test_value_modifier(self):
        factory = ResourceInstanceFactory(con=self._connection, project='test')
        SetterTester = factory.createObjectInstance('SetterTester')
        obj1 = SetterTester(stringSetter="This is a test string",
                            langStringSetter=LangString("C'est un teste@fr", "Dies ist eine Test-Zeichenkette@de"),
                            decimalSetter=Xsd_decimal(3.14),
                            integerSetter={-10, 20},
                            booleanSetter=True,
                            grantsPermission={Iri('oldap:GenericView'), Iri('oldap:GenericUpdate')})
        obj1.create()
        obj1 = SetterTester.read(con=self._connection, project='test', iri=obj1.iri)
        obj1.langStringSetter[Language.FR] = "Qu'est-ce que c'est?"
        obj1.integerSetter.add(42)
        obj1.integerSetter.discard(-10)
        obj1.booleanSetter = False
        with self.assertRaises(OldapErrorValue):
            obj1.stringSetter.pop()
        with self.assertRaises(OldapErrorValue):
            del obj1.stringSetter
        obj1.update()
        obj1 = SetterTester.read(con=self._connection, project='test', iri=obj1.iri)
        self.assertEqual(obj1.stringSetter, {"This is a test string"})
        self.assertEqual(obj1.langStringSetter, LangString("Dies ist eine Test-Zeichenkette@de", "Qu'est-ce que c'est?@fr"))
        self.assertEqual(obj1.integerSetter, {Xsd_int(20), Xsd_int(42)})
        self.assertFalse(obj1.booleanSetter)

    def test_value_modifier_norights(self):
        # ps = PermissionSet(con=self._connection,
        #                    permissionSetId="testNoUpdate",
        #                    label=LangString("testNoUpdate@en"),
        #                    comment=LangString("Testing PermissionSet@en"),
        #                    givesPermission=DataPermission.DATA_VIEW,
        #                    definedByProject="test")
        # ps.create()
        #
        # user = User(con=self._connection,
        #             userId=Xsd_NCName("factorytestuser"),
        #             familyName="FactoryTest",
        #             givenName="FactoryTest",
        #             credentials="Waseliwas",
        #             inProject={'oldap:Test': {}},
        #             hasPermissions={ps.iri.as_qname},
        #             isActive=True)
        # user.create()

        factory = ResourceInstanceFactory(con=self._connection, project='test')
        SetterTester = factory.createObjectInstance('SetterTester')
        obj1 = SetterTester(stringSetter="This is a test string",
                            langStringSetter=LangString("C'est un teste@fr", "Dies ist eine Test-Zeichenkette@de"),
                            decimalSetter={Xsd_decimal(3.14159), Xsd_decimal(2.71828), Xsd_decimal(1.61803)},
                            integerSetter={-10, 20},
                            booleanSetter=True,
                            grantsPermission={Iri('oldap:GenericView'), Iri('oldap:GenericUpdate'), self._tps.iri})
        obj1.create()

        unpriv = Connection(server='http://localhost:7200',
                            repo="oldap",
                            userId="factorytestuser",
                            credentials="Waseliwas",
                            context_name="DEFAULT")
        factory = ResourceInstanceFactory(con=unpriv, project='test')
        SetterTester = factory.createObjectInstance('SetterTester')
        obj = SetterTester.read(con=unpriv, project='test', iri=obj1.iri)
        self.assertEqual(obj.stringSetter, {"This is a test string"})
        self.assertEqual(obj.langStringSetter, LangString("C'est un teste@fr", "Dies ist eine Test-Zeichenkette@de"))
        self.assertTrue(obj.booleanSetter)
        self.assertEqual(obj.decimalSetter, {Xsd_decimal(3.14159), Xsd_decimal(2.71828), Xsd_decimal(1.61803)})
        self.assertEqual(obj.integerSetter, {-10, 20})

        obj.decimalSetter.discard(Xsd_decimal(3.14159))
        with self.assertRaises(OldapErrorNoPermission):
            obj.update()

    def test_delete_resource(self):
        project = Project.read(con=self._connection, projectIri_SName='test')
        factory = ResourceInstanceFactory(con=self._connection, project=project)
        Book = factory.createObjectInstance('Book')
        b = Book(title="Hitchhiker's Guide to the Galaxy",
                 author=Iri('test:DouglasAdams', validate=False),
                 pubDate="1995-09-27",
                 grantsPermission=Iri('oldap:GenericView'))
        b.create()
        b = Book.read(con=self._connection,
                       project='test',
                       iri=b.iri)
        Page = factory.createObjectInstance('Page')
        p1 = Page(pageDesignation="Cover",
                  pageNum=1,
                  pageDescription=LangString("Cover page of book@en", "Vorderseite des Bucheinschlags@de"),
                  pageInBook=b.iri,
                  grantsPermission=Iri('oldap:GenericView'))
        p1.create()

        with self.assertRaises(OldapErrorInUse):
            b.delete()

        p1.delete()
        with self.assertRaises(OldapErrorNotFound):
            Page.read(con=self._connection,
                    project='test',
                    iri=p1.iri)
        b.delete()
        with self.assertRaises(OldapErrorNotFound):
            b = Book.read(con=self._connection,
                          project='test',
                          iri=b.iri)

    def test_change_permissions(self):
        project = Project.read(con=self._connection, projectIri_SName='test')

        ps = PermissionSet(con=self._connection,
                           permissionSetId="testChangePermission",
                           label=LangString("testChangePermission@en"),
                           comment=LangString("Testing PermissionSet@en"),
                           givesPermission=DataPermission.DATA_PERMISSIONS,
                           definedByProject="test")
        ps.create()

        factory = ResourceInstanceFactory(con=self._connection, project=project)
        Book = factory.createObjectInstance('Book')

        b = Book(title="Hitchhiker's Guide to the Galaxy",
                 author=Iri('test:DouglasAdams', validate=False),
                 pubDate="1995-09-27",
                 grantsPermission=Iri('oldap:GenericView'))
        b.create()
        b = Book.read(con=self._connection,
                      project='test',
                      iri=b.iri)
        b.grantsPermission.add('hyha:HyperHamletMember')
        b.update()

        unpriv = Connection(server='http://localhost:7200',
                            repo="oldap",
                            userId="factorytestuser",
                            credentials="Waseliwas",
                            context_name="DEFAULT")
        factory = ResourceInstanceFactory(con=unpriv, project=project)
        Book = factory.createObjectInstance('Book')
        b = Book.read(con=self._connection,
                      project='test',
                      iri=b.iri)
        b.grantsPermission.add('oldap:GenericUpdate')
        with self.assertRaises(OldapErrorNoPermission):
            b.update()


if __name__ == '__main__':
    unittest.main()



from copy import deepcopy
from functools import partial

from typing import List, Self, Any
from datetime import date, datetime

from oldaplib.src.cachesingleton import CacheSingleton
from oldaplib.src.enums.adminpermissions import AdminPermission
from oldaplib.src.enums.projectattr import ProjectAttr
from oldaplib.src.helpers.context import Context
from oldaplib.src.enums.action import Action
from oldaplib.src.dtypes.namespaceiri import NamespaceIRI
from oldaplib.src.helpers.irincname import IriOrNCName
from oldaplib.src.helpers.tools import lprint
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_anyuri import Xsd_anyURI
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_date import Xsd_date
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.helpers.oldaperror import OldapError, OldapErrorAlreadyExists, OldapErrorNoPermission, \
    OldapErrorUpdateFailed, OldapErrorNotFound, OldapErrorInconsistency
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.iconnection import IConnection
from oldaplib.src.model import Model
from oldaplib.src.helpers.attributechange import AttributeChange
from oldaplib.src.xsd.xsd_string import Xsd_string


#@strict
class Project(Model):
    """
    # Project

    This class implements the Project model. A Project is a distinct research space within Oldap
    framework that offers dedicated space for data, its own data modeling and access control. A project
    needs the following metadata:

    - `projectIri: Iri | None` [optional]:
      The IRI that uniquely identifies this project. This can be an
      [Iri](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.Iri) instance or a string with either an Iri or Qname
      [QName](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.QName). If the `projectIri` is ommitted,
      a random IRI based on the URN semantics will be generated.
    - `projectShortName: Xsd_NCName | str` [mandatory]:
      The short name of the project that must be a NCName or a string
    - `label: LangString | None` [optional]:
      A multilingual string with a human-readable label for the project (`rdfs:label`)
    - `comment: LangString | None` [optional]:
      A multilingual description of the project (`rdfs:comment`)
    - `namespaceIri: NamespaceIri` [mandatory]:
       The namespace that the project uses for its data and data model. Must be
       a [NamespaceIRI](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.NamespaceIRI).
    - `projectStart: Xsd_date | date | None` [optional]:
      The start date of the project.  Must be a Python `date` type. If not set,
       the current date will be used.
    - `projectEnd: Xsd_date | None` [optional]:
      The optional end date of the project. Must be a Xsd_date or a Python `date` type.

    The class provides the following methods:

    - [Project(...)](/python_docstrings/project#oldaplib.src.project.Project.__init__)): Constructor
    - [read(...)](/python_docstrings/project#oldaplib.src.project.Project.read):
      Read project form triplestore and return a Project-instance
    - [search(...)](/python_docstrings/project#oldaplib.src.project.Project.search):
      Search a Project in the triplestore
    - [create(...)](/python_docstrings/project#oldaplib.src.project.Project.create): Create a new project in the triplestore
    - [update(...)](/python_docstrings/project#oldaplib.src.project.Project.update):
      Write changes to triplestore
    - [delete(...)](/python_docstrings/project#oldaplib.src.project.Project.update):
      Delete a project in the triplestore

    The class provides the following properties:

    - `projectIri` [read only]: The project IRI
    - `projectShortName` [read/write]: The project short name. Must be a NCName
    - `projectLabel` [read/write]: The projects label as multilingual LangString
    - `projectComment` [read/write]: The projects comment/description as multilingual
      [LangString](/python_docstrings/langstring)
    - `namespaceIri` [read only]: The namespace that the project uses for its data and data model. Must be
      a [NamespaceIRI](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.NamespaceIRI).
    - `projectStart` [read/write]: The start date of the project. Must be a Python `date` type
    - `projectEnd` [read/write]: The end date of the project. Must be a Python `date` type
    - `creator` [read only]: The creator of the project. Must be a
      [AnyIRI](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.AnyIRI) or
      [QName](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.AnyIRI)
    - `created` [read only]: The creation date of the project. Must be a Python `date` type
    - `contributor` [read only]: The person which made the last changes to the project data. Must be a
      [AnyIRI](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.AnyIRI) or
      [QName](/python_docstrings/datatypes#oldaplib.src.helpers.datatypes.AnyIRI)
    - `modified` [read only]: The modification date of the project. Must be a Python `date` type

    """

    #_attributes: dict[ProjectAttr, ProjectAttrTypes]
    #__changeset: dict[ProjectAttr, ProjectAttrChange]

    def __init__(self, *,
                 con: IConnection,
                 creator: Iri | None = None,
                 created: Xsd_dateTime | None = None,
                 contributor: Iri | None = None,
                 modified: Xsd_dateTime | None = None,
                 **kwargs):
        """
        Constructs a new Project
        :param con: [Connection](/python_docstrings/iconnection) instance
        :param creator: Creator of the project  [DO NOT SET!]
        :type creator: Xsd_anyURI | None
        :param created: Date the project was created  [DO NOT SET!]
        :type created: datetime | None
        :param contributor: person that made the last change  [DO NOT SET!]
        :type contributor: Xsd_anyURI
        :param modified: Last date the project was modified  [DO NOT SET!]
        :type modified: date | None
        :param projectIri: IRI to be used for the project. If no projectIRI is provied, the constrctor
         will create an arbitrary IRI based on thr URN scheme and a UUID. [optional].
        :type projectIri: Iri
        :param projectShortName: A short name for the project. Is used as prefix for named graphs that
           are being used for the project. [mandatory]
        :type projectShortName: NCname (strings are accepted only if conform to NCName syntax)
        :param namespaceIri: The namespace to be used for the projects data [mandatory]
        :type namespaceIri: NamespaceIRI
        :param label: Human-readable name for the project (multi-language) (`rdfs:label`) [optional]
        :type label: LangString
        :param comment: Description of the project (multi-language) (`rdfs:comment`)
        :type comment: LangString
        :param projectStart: Start date of the project
        :type projectStart: Python date
        :param projectEnd: End date of the project [Optional]
        :type projectEnd: Python date
        :raises OldapErrorValue: Invalid parameter supplied
        """
        super().__init__(connection=con,
                         created=created,
                         creator=creator,
                         modified=modified,
                         contributor=contributor)

        self.set_attributes(kwargs, ProjectAttr)
        #
        # Consistency checks
        #
        if not self._attributes.get(ProjectAttr.PROJECT_IRI):
            self._attributes[ProjectAttr.PROJECT_IRI] = Iri()
        if not self._attributes.get(ProjectAttr.PROJECT_START):
            self._attributes[ProjectAttr.PROJECT_START] = Xsd_date()
        if self._attributes.get(ProjectAttr.PROJECT_END):
            self.check_consistency(ProjectAttr.PROJECT_END, self._attributes[ProjectAttr.PROJECT_END])
        if self._attributes.get(ProjectAttr.PROJECT_START):
            self.check_consistency(ProjectAttr.PROJECT_START, self._attributes[ProjectAttr.PROJECT_START])
        #
        # create all the attributes of the class according to the ProjectFields dfinition
        #
        for attr in ProjectAttr:
            setattr(Project, attr.value.fragment, property(
                partial(Project._get_value, attr=attr),
                partial(Project._set_value, attr=attr),
                partial(Project._del_value, attr=attr)))
        self._changeset = {}

    def __deepcopy__(self, memo: dict[Any, Any]) -> Self:
        if id(self) in memo:
            return memo[id(self)]
        cls = self.__class__
        instance = cls.__new__(cls)
        memo[id(self)] = instance
        Model.__init__(instance,
                       connection=deepcopy(self._con, memo),
                       creator=deepcopy(self._creator, memo),
                       created=deepcopy(self._created, memo),
                       contributor=deepcopy(self._contributor, memo),
                       modified=deepcopy(self._modified, memo))
        # Copy internals of Model:
        instance._attributes = deepcopy(self._attributes, memo)
        instance._changset = deepcopy(self._changeset, memo)
        return instance


    def check_consistency(self, attr: ProjectAttr, value: Any) -> None:
        if attr == ProjectAttr.PROJECT_END:
            if self._attributes.get(ProjectAttr.PROJECT_START) and value < self._attributes[ProjectAttr.PROJECT_START]:
                raise OldapErrorInconsistency(f'Project start date {self._attributes[ProjectAttr.PROJECT_START]} is after project end date {value}.')

        if attr == ProjectAttr.PROJECT_START:
            if self._attributes.get(ProjectAttr.PROJECT_END) and value > self._attributes[ProjectAttr.PROJECT_END]:
                raise OldapErrorInconsistency(f'Project start date {value} is after project end date {self._attributes[ProjectAttr.PROJECT_END]}.')

    def check_for_permissions(self) -> (bool, str):
        actor = self._con.userdata
        sysperms = actor.inProject.get(Iri('oldap:SystemProject'))
        if sysperms and AdminPermission.ADMIN_OLDAP in sysperms:
            return True, "OK"
        else:
            return False, "No permission to create/change a project."

    def notifier(self, attr: ProjectAttr) -> None:
        """
        This method is called when a field is being changed.
        :param fieldname: Fieldname of the field being modified
        :return: None
        """
        self._changeset[attr] = AttributeChange(self._attributes[attr], Action.MODIFY)

    @classmethod
    def read(cls,
             con: IConnection,
             projectIri_SName: IriOrNCName | Iri | Xsd_NCName | str,
             ignore_cache: bool = False) -> Self:
        """
        Read the project from the triplestore and return an instance of the project
        :param con: A valid Connection object
        :type con: IConnection
        :param projectIri_SName: The IRI or shortname of the project to be read
        :type projectIri_SName: Iri | Xsd_NCName | str
        :return: Project instance
        :rtype: Project
        :raise: OldapErrorNotFound: project with given Iri not found
        :raise: OldapError: All other errors/problems
        """
        context = Context(name=con.context_name)
        query = context.sparql_context

        if not isinstance(projectIri_SName, IriOrNCName):
            projectIri_SName = IriOrNCName(projectIri_SName)
        shortname, projectIri = projectIri_SName.value()
        # projectIri: Iri | None = None
        # shortname: Xsd_NCName | None = None
        # if isinstance(projectIri_SName, Iri):
        #     projectIri = projectIri_SName
        # elif isinstance(projectIri_SName, Xsd_NCName):
        #     shortname = Xsd_NCName(projectIri_SName)
        # else:
        #     if ':' in str(projectIri_SName):  # must be IRI or QName
        #         projectIri = Iri(projectIri_SName)
        #     else:
        #         shortname = Xsd_NCName(projectIri_SName)
        cache = CacheSingleton()
        if projectIri is not None:
            if not ignore_cache:
                tmp = cache.get(projectIri)
                if tmp is not None:
                    tmp._con = con
                    return tmp
            query += f"""
                SELECT ?prop ?val
                FROM NAMED oldap:admin
                WHERE {{
                    GRAPH oldap:admin {{
                        {projectIri.toRdf} ?prop ?val
                    }}
                }}
            """
        elif shortname is not None:
            if not ignore_cache:
                tmp = cache.get(shortname)
                if tmp is not None:
                    tmp._con = con
                    return tmp
            query += f"""
                SELECT ?proj ?prop ?val
                FROM NAMED oldap:admin
                WHERE {{
                    GRAPH oldap:admin {{
                        ?proj a oldap:Project .
                        ?proj oldap:projectShortName ?shortname .
                        ?proj ?prop ?val .
                    }}
                    FILTER(?shortname = {shortname.toRdf})
                }}
            """
        jsonobj = con.query(query)
        res = QueryProcessor(context, jsonobj)
        if len(res) == 0:
            raise OldapErrorNotFound(f'Project with IRI/shortname "{projectIri_SName}" not found.')
        creator: Iri | None = None
        created: Xsd_dateTime | None = None
        contributor: Iri | None = None
        modified: Xsd_dateTime | None = None
        projectShortName = None
        namespaceIri = None
        label = LangString()
        comment = LangString()
        projectStart = None
        projectEnd = None
        for r in res:
            if projectIri is None:
                projectIri = r['proj']
            match str(r.get('prop')):
                case 'dcterms:creator':
                    creator = r['val']
                case 'dcterms:created':
                    created = r['val']
                case 'dcterms:contributor':
                    contributor = r['val']
                case 'dcterms:modified':
                    modified = r['val']
                case 'oldap:namespaceIri':
                    namespaceIri = NamespaceIRI(r['val'])
                case 'oldap:projectShortName':
                    projectShortName = r['val']
                case 'rdfs:label':
                    label.add(str(r['val']))
                case 'rdfs:comment':
                    comment.add(str(r['val']))
                case 'oldap:projectStart':
                    projectStart = r['val']
                case 'oldap:projectEnd':
                    projectEnd = r['val']
        if label:
            label.changeset_clear()
            label.set_notifier(cls.notifier, Xsd_QName(ProjectAttr.LABEL.value))
        if comment:
            comment.changeset_clear()
            comment.set_notifier(cls.notifier, Xsd_QName(ProjectAttr.COMMENT.value))
        context[projectShortName] = namespaceIri
        instance = cls(con=con,
                       creator=creator,
                       created=created,
                       contributor=contributor,
                       modified=modified,
                       projectIri=projectIri,
                       projectShortName=projectShortName,
                       label=label,
                       namespaceIri=namespaceIri,
                       comment=comment,
                       projectStart=projectStart,
                       projectEnd=projectEnd)
        cache = CacheSingleton()
        cache.set(instance.projectIri, instance, instance.projectShortName)
        return instance

    @staticmethod
    def search(con: IConnection,
               label: Xsd_string | str | None = None,
               comment: Xsd_string | str | None = None) -> list[Iri]:
        """
        Search for a given project. If no label or comment is given, all existing projects are returned. If both
        a search term for the label and comment are given, they will be combined by *AND*.
        :param con: Valid Connection object
        :type con: IConnection
        :param label: A string to search for in the labels. The search will check if the label of a project
        **contains** the string given here.
        :type label: str
        :param comment: A string to search for in the comments. The search will check if the comment of a project
        **contains** the string given here
        :type comment: str
        :return: List of IRIs matching the search criteria (AnyIRI | QName)
        :rtype: List[Iri]
        :raises OldapErrorNotFound: If the project does not exist
        """
        label = Xsd_string(label)
        comment = Xsd_string(comment)
        context = Context(name=con.context_name)
        sparql = context.sparql_context
        sparql += 'SELECT DISTINCT ?project\n'
        sparql += 'FROM oldap:onto\n'
        sparql += 'FROM shared:onto\n'
        sparql += 'FROM NAMED oldap:admin\n'
        sparql += 'WHERE {\n'
        sparql += '    GRAPH oldap:admin {\n'
        sparql += '        ?project a oldap:Project .\n'
        if label:
            sparql += '        ?project rdfs:label ?label .\n'
            sparql += '    }\n'
            sparql += f'    FILTER(CONTAINS(STR(?label), "{Xsd_string.escaping(label.value)}"))\n'
        if comment:
            sparql += '        ?project rdfs:comment ?comment .\n'
            sparql += '    }\n'
            sparql += f'    FILTER(CONTAINS(STR(?comment), "{Xsd_string.escaping(comment.value)}"))\n'
        sparql += '}\n'
        jsonobj = con.query(sparql)
        res = QueryProcessor(context, jsonobj)
        projects: list[Iri] = []
        if len(res) > 0:
            for r in res:
                projects.append(r['project'])
        return projects

    def create(self, indent: int = 0, indent_inc: int = 4) -> None:
        """
        Create a new project in the triple store
        :param indent: Start indent level for generated SPARQL (debugging)
        :type indent: int
        :param indent_inc: Indent increment
        :type indent_inc: int
        :return: None
        :raises OldapErrorAlreadyExists: If a project with the projectIri already exists
        :raises OldapError: All other errors
        """
        if self._con is None:
            raise OldapError("Cannot create: no connection")
        #
        # First we check if the logged-in user ("actor") has the permission to create a user for
        # the given project!
        #
        result, message = self.check_for_permissions()
        if not result:
            raise OldapErrorNoPermission(message)

        timestamp = Xsd_dateTime.now()
        indent: int = 0
        indent_inc: int = 4

        context = Context(name=self._con.context_name)

        sparql1 = context.sparql_context
        sparql1 += f"""
        SELECT ?project
        FROM oldap:admin
        WHERE {{
            ?project a oldap:Project .
            FILTER(?project = {self.projectIri.toRdf})
        }}
        """

        blank = ''
        sparql2 = context.sparql_context
        sparql2 += f'{blank:{indent * indent_inc}}INSERT DATA {{'
        sparql2 += f'\n{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{'
        sparql2 += f'\n{blank:{(indent + 2) * indent_inc}}{self.projectIri.toRdf} a oldap:Project'
        sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:creator {self._con.userIri.toRdf}'
        sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:created {timestamp.toRdf}'
        sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:contributor {self._con.userIri.toRdf}'
        sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:modified {timestamp.toRdf}'
        for attr, value in self._attributes.items():
            if not value:
                continue
            sparql2 += f' ;\n{blank:{(indent + 3) * indent_inc}}{attr.value.toRdf} {value.toRdf}'
        sparql2 += f' .\n{blank:{(indent + 1) * indent_inc}}}}\n'
        sparql2 += f'{blank:{indent * indent_inc}}}}\n'

        self._con.transaction_start()
        try:
            jsonobj = self._con.transaction_query(sparql1)
        except OldapError:
            self._con.transaction_abort()
            raise
        res = QueryProcessor(context, jsonobj)
        if len(res) > 0:
            self._con.transaction_abort()
            raise OldapErrorAlreadyExists(f'A Project with a projectIri "{self.projectIri}" already exists')

        try:
            self._con.transaction_update(sparql2)
        except OldapError:
            self._con.transaction_abort()
            raise
        try:
            self._con.transaction_commit()
        except OldapError:
            self._con.transaction_abort()
            raise
        self._created = timestamp
        self._creator = self._con.userIri
        self._modified = timestamp
        self._contributor = self._con.userIri
        context[self._attributes[ProjectAttr.PROJECT_SHORTNAME]] = self._attributes[ProjectAttr.NAMESPACE_IRI]

        cache = CacheSingleton()
        cache.set(self.projectIri, self, self.projectShortName)

    def update(self, indent: int = 0, indent_inc: int = 4) -> None:
        """
        Write all the modifications that were applied to the project instqnce to the triple store.
        :param indent: Starting indent for SPARQL queries [Only used for debbugging purposes]
        :type indent: int
        :param indent_inc: Indent increment for SPARQL queries [Only used for debbugging purposes]
        :type indent_inc: int
        :return: None
        :Raises: OldapErrorNoPermission: No permission for operation
        :Raises: OldapErrorUpdateFailed: Update failed
        :Raises: Oldap Error: Other Internal error
        """
        result, message = self.check_for_permissions()
        if not result:
            raise OldapErrorNoPermission(message)

        timestamp = Xsd_dateTime.now()
        context = Context(name=self._con.context_name)
        blank = ''
        sparql_list = []
        for field, change in self._changeset.items():
            if field == ProjectAttr.LABEL or field == ProjectAttr.COMMENT:
                if change.action == Action.MODIFY:
                    sparql_list.extend(self._attributes[field].update(graph=Xsd_QName('oldap:admin'),
                                                                      subject=self.projectIri,
                                                                      field=Xsd_QName(field.value)))
                if change.action == Action.DELETE or change.action == Action.REPLACE:
                    sparql = self._changeset[field].old_value.delete(graph=Xsd_QName('oldap:admin'),
                                                                      subject=self.projectIri,
                                                                      field=Xsd_QName(field.value))
                    sparql_list.append(sparql)
                if change.action == Action.CREATE or change.action == Action.REPLACE:
                    sparql = self._attributes[field].create(graph=Xsd_QName('oldap:admin'),
                                                            subject=self.projectIri,
                                                            field=Xsd_QName(field.value))
                    sparql_list.append(sparql)
                continue
            sparql = f'{blank:{indent * indent_inc}}# Project field "{field.value}" with action "{change.action.value}"\n'
            sparql += f'{blank:{indent * indent_inc}}WITH oldap:admin\n'
            if change.action != Action.CREATE:
                sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?project {field.value} {change.old_value.toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            if change.action != Action.DELETE:
                sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?project {field.value} {self._attributes[field].toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}BIND({self.projectIri.toRdf} as ?project)\n'
            if change.action != Action.CREATE:
                sparql += f'{blank:{(indent + 1) * indent_inc}}?project {field.value} {change.old_value.toRdf} .\n'
            sparql += f'{blank:{indent * indent_inc}}}}'
            sparql_list.append(sparql)
        sparql = context.sparql_context
        sparql += " ;\n".join(sparql_list)

        self._con.transaction_start()
        try:
            self._con.transaction_update(sparql)
            self.set_modified_by_iri(Xsd_QName('oldap:admin'), self.projectIri, self.modified, timestamp)
            modtime = self.get_modified_by_iri(Xsd_QName('oldap:admin'), self.projectIri)
        except OldapError:
            self._con.transaction_abort()
            raise
        if timestamp != modtime:
            self._con.transaction_abort()
            raise OldapErrorUpdateFailed("Update failed! Timestamp does not match")
        try:
            self._con.transaction_commit()
        except OldapError:
            self._con.transaction_abort()
            raise
        self._modified = timestamp
        self._contributor = self._con.userIri
        self.clear_changeset()
        cache = CacheSingleton()
        cache.set(self.projectIri, self, self.projectShortName)

    def delete(self) -> None:
        """
        Delete the given user from the triplestore
        :return: None
        :raises OldapErrorNoPermission: No permission for operation
        :raises OldapError: generic internal error
        """
        result, message = self.check_for_permissions()
        if not result:
            raise OldapErrorNoPermission(message)

        #
        # TODO: Check if project has any datamodel and/or data. Decline the deletion if this is the case
        #
        context = Context(name=self._con.context_name)
        sparql = context.sparql_context
        sparql += f"""
        DELETE WHERE {{
            GRAPH oldap:admin {{
                {self.projectIri.toRdf} a oldap:Project .
                {self.projectIri.toRdf} ?prop ?val .
            }}
        }} 
        """
        self._con.update_query(sparql)
        cache = CacheSingleton()
        cache.delete(self.projectIri)
        cache.delete(self.projectShortName)





from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName


class IriOrNCName:

    def __init__(self, value: Iri | Xsd_NCName | str):
        self.__iri: Iri | None = None
        self.__ncname: Xsd_NCName | None = None
        if isinstance(value, Iri):
            self.__iri = value
            self.__ncname = None
        elif isinstance(value, Xsd_NCName):
            self.__ncname = Xsd_NCName(value)
            self.__iri = None
        else:
            if ':' in str(value):  # must be IRI or QName
                self.__iri = Iri(value)
                self.__ncname = None
            else:
                self.__ncname = Xsd_NCName(value)
                self.__iri = None

    def value(self) -> tuple[Xsd_NCName| None, Iri | None]:
        return self.__ncname, self.__iri

    def __str__(self):
        if self.__iri is not None:
            return str(self.__iri)
        elif self.__ncname is not None:
            return str(self.__ncname)
        else:
            return "???"



from copy import deepcopy
from dataclasses import dataclass
from functools import partial
from typing import Callable, Self, Any

from oldaplib.src.enums.action import Action
from oldaplib.src.enums.haspropertyattr import HasPropertyAttr
from oldaplib.src.helpers.Notify import Notify
from oldaplib.src.iconnection import IConnection
from oldaplib.src.model import Model
from oldaplib.src.helpers.attributechange import AttributeChange
from oldaplib.src.propertyclass import PropertyClass, HasPropertyData
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_decimal import Xsd_decimal
from oldaplib.src.xsd.xsd_integer import Xsd_integer
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_nonnegativeinteger import Xsd_nonNegativeInteger


class HasProperty(Model, Notify):
    _prop: PropertyClass | Iri | None

    def __init__(self, *,
                 con: IConnection,
                 prop: PropertyClass | Iri | None = None,
                 notifier: Callable[[Iri], None] | None = None,
                 notify_data: Iri | None = None,
                 **kwargs):
        Model.__init__(self, connection=con)
        Notify.__init__(self, notifier, notify_data)
        self._prop = prop
        self.set_attributes(kwargs, HasPropertyAttr)

        for attr in HasPropertyAttr:
            setattr(HasProperty, attr.value.fragment, property(
                partial(HasProperty._get_value, attr=attr),
                partial(HasProperty._set_value, attr=attr),
                partial(HasProperty._del_value, attr=attr)))
        self._changeset = {}

    def __deepcopy__(self, memo: dict[Any, Any]) -> Self:
        if id(self) in memo:
            return memo[id(self)]
        cls = self.__class__
        instance = cls.__new__(cls)
        memo[id(self)] = instance
        Model.__init__(instance,
                       connection=deepcopy(self._con, memo),
                       creator=deepcopy(self._creator, memo),
                       created=deepcopy(self._created, memo),
                       contributor=deepcopy(self._contributor, memo),
                       modified=deepcopy(self._modified, memo))
        Notify.__init__(instance,
                        notifier=self._notifier,
                        data=deepcopy(self._notify_data, memo))
        # Copy internals of Model:
        instance._attributes = deepcopy(self._attributes, memo)
        instance._changset = deepcopy(self._changeset, memo)
        instance._prop = deepcopy(self._prop)  # no deepcopy here (????)
        return instance

    def __str__(self) -> str:
        iri = self._prop.property_class_iri if isinstance(self._prop, PropertyClass) else self._prop
        s = f'HasProperty: {iri}\n'
        s += Model.__str__(self)
        return s

    @property
    def prop(self) -> PropertyClass:
        return self._prop

    @prop.setter
    def prop(self, prop: PropertyClass) -> None:
        self._prop = prop

    @property
    def haspropdata(self) -> HasPropertyData:
        return HasPropertyData(minCount=self._attributes.get(HasPropertyAttr.MIN_COUNT, None),
                               maxCount=self._attributes.get(HasPropertyAttr.MAX_COUNT, None),
                               order=self._attributes.get(HasPropertyAttr.ORDER, None),
                               group=self._attributes.get(HasPropertyAttr.GROUP, None))

    def notifier(self, attr: HasPropertyAttr) -> None:
        if attr == HasPropertyAttr.PROP:
            return
        self._changeset[attr] = AttributeChange(self._attributes[attr], Action.MODIFY)
        self.notify()

    def create_shacl(self, indent: int = 0, indent_inc: int = 4) -> str:
        blank = ''
        sparql = ''
        if self._attributes.get(HasPropertyAttr.MIN_COUNT, None) is not None:
            sparql += f' ;\n{blank:{indent * indent_inc}}sh:minCount {self._attributes[HasPropertyAttr.MIN_COUNT].toRdf}'
        if self._attributes.get(HasPropertyAttr.MAX_COUNT, None) is not None:
            sparql += f' ;\n{blank:{indent * indent_inc}}sh:maxCount {self._attributes[HasPropertyAttr.MAX_COUNT].toRdf}'
        if self._attributes.get(HasPropertyAttr.ORDER, None) is not None:
            sparql += f' ;\n{blank:{indent * indent_inc}}sh:order {self._attributes[HasPropertyAttr.ORDER].toRdf}'
        if self._attributes.get(HasPropertyAttr.GROUP, None) is not None:
            sparql += f' ;\n{blank:{indent * indent_inc}}sh:group {self._attributes[HasPropertyAttr.GROUP].toRdf}'
        return sparql

    def create_owl(self, indent: int = 0, indent_inc: int = 4):
        blank = ''
        sparql = ''
        min_count = Xsd_nonNegativeInteger(int(self._attributes[HasPropertyAttr.MIN_COUNT])) if self._attributes.get(HasPropertyAttr.MIN_COUNT) else None
        max_count = Xsd_nonNegativeInteger(int(self._attributes[HasPropertyAttr.MAX_COUNT])) if self._attributes.get(HasPropertyAttr.MAX_COUNT) else None

        if min_count and max_count and min_count == max_count:
            sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}owl:qualifiedCardinality {min_count.toRdf}'
        else:
            if min_count:
                sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}owl:minQualifiedCardinality {min_count.toRdf}'
            if max_count:
                sparql += f' ;\n{blank:{(indent + 1) * indent_inc}}owl:maxQualifiedCardinality {max_count.toRdf}'
        return sparql

    def update_shacl(self,
                     graph: Xsd_NCName,
                     resclass_iri: Iri,
                     propclass_iri: Iri,
                     indent: int = 0, indent_inc: int = 4) -> str:
        blank = ''
        sparql_list = []
        for attr, change in self._changeset.items():
            sparql = f'WITH {graph}:shacl\n'
            if change.action != Action.CREATE:
                sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?prop {attr.value} {change.old_value.toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'

            if change.action != Action.DELETE:
                sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?prop {attr.value} {self._attributes[attr].toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'

            sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}{resclass_iri}Shape sh:property ?prop .\n'
            if isinstance(self._prop, Iri) or self._prop.internal is None:
                sparql += f'{blank:{(indent + 1) * indent_inc}}?prop sh:node {propclass_iri}Shape.\n'
            else:
                sparql += f'{blank:{(indent + 1) * indent_inc}}?prop sh:path {propclass_iri} .\n'
            if change.action != Action.CREATE:
                sparql += f'{blank:{(indent + 1) * indent_inc}}?prop {attr.value} {change.old_value.toRdf} .\n'
            sparql += f'{blank:{indent * indent_inc}}}}'
            sparql_list.append(sparql)
        sparql = " ;\n".join(sparql_list)
        return sparql

    def update_owl(self,
                     graph: Xsd_NCName,
                     resclass_iri: Iri,
                     propclass_iri: Iri,
                     indent: int = 0, indent_inc: int = 4) -> str:
        blank = ''
        if HasPropertyAttr.MIN_COUNT in self._changeset or HasPropertyAttr.MAX_COUNT in self._changeset:
            sparql = f'WITH {graph}:onto\n'
            sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}?prop owl:qualifiedCardinality ?val_qualified .\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}?prop owl:minQualifiedCardinality ?val_min .\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}?prop owl:maxQualifiedCardinality ?val .\n'
            sparql += f'{blank:{indent * indent_inc}}}}\n'

            #
            # Convert to xsd:nonNegativeInteger or None
            #
            min_count = Xsd_nonNegativeInteger(int(self._attributes[HasPropertyAttr.MIN_COUNT])) if self._attributes.get(HasPropertyAttr.MIN_COUNT) else None
            max_count = Xsd_nonNegativeInteger(int(self._attributes[HasPropertyAttr.MAX_COUNT])) if self._attributes.get(HasPropertyAttr.MAX_COUNT) else None

            if min_count or max_count:
                sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
                if min_count and max_count and min_count == max_count:
                    sparql += f'{blank:{(indent + 1) * indent_inc}}?prop owl:qualifiedCardinality {min_count.toRdf} .\n'
                else:
                    if min_count:
                        sparql += f'{blank:{(indent + 1) * indent_inc}}?prop owl:minQualifiedCardinality {min_count.toRdf} .\n'
                    if max_count:
                        sparql += f'{blank:{(indent + 1) * indent_inc}}?prop owl:maxQualifiedCardinality {max_count.toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'

            sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}{resclass_iri} rdfs:subClassOf ?prop .\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}?prop owl:onProperty {propclass_iri} .\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}OPTIONAL {{ ?prop owl:qualifiedCardinality ?val_qualified . }}\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}OPTIONAL {{ ?prop owl:minQualifiedCardinality ?val_min . }}\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}OPTIONAL {{ ?prop owl:maxQualifiedCardinality ?val_max . }}\n'
            sparql += f'{blank:{indent * indent_inc}}}}'

            return sparql



from enum import unique, Enum

from oldaplib.src.enums.attributeclass import AttributeClass
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.helpers.observable_dict import ObservableDict
from oldaplib.src.xsd.xsd_boolean import Xsd_boolean
from oldaplib.src.xsd.xsd_qname import Xsd_QName


@unique
class ResClassAttribute(AttributeClass):
    # order: (QName, mandatory, immutable, datatype)
    SUPERCLASS = ('oldap:superclass', False, False, ObservableDict)  # virtual attribute, SHACL: sh:node, OWL: rdfs:subClassOf
    LABEL = ('rdfs:label', False, False, LangString)
    COMMENT = ('rdfs:comment', False, False, LangString)
    CLOSED = ('sh:closed', False, False, Xsd_boolean)




from copy import deepcopy
from threading import Lock
from typing import Any

from oldaplib.src.helpers.singletonmeta import SingletonMeta
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName


class CacheSingleton(metaclass=SingletonMeta):
    _lock: Lock
    _cache: dict[Iri | Xsd_NCName, Any]

    def __init__(self):
        self._lock = Lock()
        self._cache = {}

    def __str__(self) -> str:
        with self._lock:
            return str(self._cache)

    def get(self, key: Iri | Xsd_NCName) -> Any:
        with self._lock:
            return deepcopy(self._cache.get(key))

    def set(self, key: Iri | Xsd_NCName, value: Any, key2: Iri | Xsd_NCName | None = None) -> None:
        with self._lock:
            self._cache[key] = deepcopy(value)
            if key2 is not None:
                self._cache[key2] = self._cache[key]

    def delete(self, key: Iri | Xsd_NCName):
        with self._lock:
            if key in self._cache:
                self._cache.pop(key, None)

    def clear(self):
        with self._lock:
            self._cache.clear()



import unittest
from copy import deepcopy
from datetime import date
from pathlib import Path
from time import sleep

from oldaplib.src.cachesingleton import CacheSingleton
from oldaplib.src.connection import Connection
from oldaplib.src.enums.language import Language
from oldaplib.src.helpers.context import Context
from oldaplib.src.dtypes.namespaceiri import NamespaceIRI
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_date import Xsd_date
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.helpers.oldaperror import OldapErrorNotFound, OldapErrorInconsistency, OldapErrorNoPermission
from oldaplib.src.project import Project


def find_project_root(current_path):
    # Climb up the directory hierarchy and check for a marker file
    path = Path(current_path).absolute()
    while not (path / 'pyproject.toml').exists():
        if path.parent == path:
            # Root of the filesystem, file not found
            raise RuntimeError('Project root not found')
        path = path.parent
    return path


class Testproject(unittest.TestCase):
    _connection: Connection
    _unpriv: Connection

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        project_root = find_project_root(__file__)
        cls._context = Context(name="DEFAULT")

        cls._connection = Connection(server='http://localhost:7200',
                                     repo="oldap",
                                     userId="rosenth",
                                     credentials="RioGrande",
                                     context_name="DEFAULT")
        cls._unpriv = Connection(server='http://localhost:7200',
                                 repo="oldap",
                                 userId="fornaro",
                                 credentials="RioGrande",
                                 context_name="DEFAULT")


        cls._connection.clear_graph(Xsd_QName('oldap:admin'))
        file = project_root / 'oldaplib' / 'ontologies' / 'admin.trig'
        cls._connection.upload_turtle(file)
        sleep(1)  # upload may take a while...

    @classmethod
    def tearDownClass(cls):
        #cls._connection.clear_graph(Xsd_QName('oldap:admin'))
        #cls._connection.upload_turtle("oldaplib/ontologies/admin.trig")
        #sleep(1)  # upload may take a while...
        pass

    def test_project_deepcopy(self):
        project = Project(con=self._connection,
                          projectShortName="unittest",
                          label=LangString(["unittest@en", "unittest@de"]),
                          namespaceIri=NamespaceIRI("http://unitest.org/project/unittest#"),
                          comment=LangString(["For testing@en", "Für Tests@de"]),
                          projectStart=Xsd_date(2024, 1, 1),
                          projectEnd=Xsd_date(2025, 12, 31)
                          )
        project2 = deepcopy(project)
        self.assertFalse(project is project2)
        self.assertEqual(project.projectIri, project2.projectIri)
        self.assertEqual(project.projectShortName, project2.projectShortName)
        self.assertEqual(project.label, project2.label)
        self.assertEqual(project.comment, project2.comment)
        self.assertEqual(project.projectStart, project2.projectStart)
        self.assertEqual(project.projectEnd, project2.projectEnd)


    # @unittest.skip('Work in progress')
    def test_project_read(self):
        project = Project.read(con=self._connection, projectIri_SName=Iri("oldap:SystemProject"), ignore_cache=True)
        self.assertEqual(Xsd_NCName("oldap"), project.projectShortName)
        self.assertEqual(LangString(["System@en",
                                     "System@de",
                                     "Système@fr",
                                     "Systema@it"]), project.label)
        self.assertEqual(NamespaceIRI("http://oldap.org/base#"), project.namespaceIri)
        self.assertEqual(LangString(["Project for system administration@en"]), project.comment)
        self.assertEqual(Xsd_date("2024-01-01"), project.projectStart)

        project = Project.read(con=self._connection, projectIri_SName='http://www.salsah.org/version/2.0/SwissBritNet', ignore_cache=True)
        self.assertEqual(Xsd_NCName("britnet"), project.projectShortName)

        project2 = Project.read(con=self._connection, projectIri_SName='britnet', ignore_cache=True)
        self.assertEqual(Xsd_NCName("britnet"), project2.projectShortName)
        self.assertEqual(Iri('http://www.salsah.org/version/2.0/SwissBritNet'), project2.projectIri)

    def test_project_read_from_cache(self):
        project = Project.read(con=self._connection, projectIri_SName='britnet', ignore_cache=True)
        project2 = Project.read(con=self._connection, projectIri_SName='britnet')
        self.assertFalse(project is project2)
        self.assertEqual(project.projectShortName, project2.projectShortName)
        self.assertFalse(project.projectShortName is project2.projectShortName)

    def test_project_read_unknown(self):
        with self.assertRaises(OldapErrorNotFound):
            project = Project.read(con=self._connection, projectIri_SName='http://www.gaga.org/version/2.0/GAGA', ignore_cache=True)


    # @unittest.skip('Work in progress')
    def test_project_search(self):
        projects = Project.search(con=self._connection, label="HyperHamlet")
        self.assertEqual(["oldap:HyperHamlet"], projects)

    # @unittest.skip('Work in progress')
    def test_project_search_fail(self):
        projects = Project.search(con=self._connection, label="NoExisting")
        self.assertEqual([], projects)

    def test_project_create_simplest(self):
        project = Project(con=self._connection,
                          projectShortName="simpleproj",
                          namespaceIri=NamespaceIRI("http://unitest.org/project/simpleproj#"),
                          projectStart=Xsd_date(2024, 1, 1),
                          projectEnd=Xsd_date(2025, 12, 31)
                          )
        project.create()
        self.assertIsNotNone(project.created)
        self.assertIsNotNone(project.creator)
        self.assertIsNotNone(project.modified)
        self.assertIsNotNone(project.contributor)
        self.assertEqual("simpleproj", project.projectShortName)
        self.assertEqual("http://unitest.org/project/simpleproj#", project.namespaceIri)

    # @unittest.skip('Work in progress')
    def test_project_create(self):
        project = Project(con=self._connection,
                          projectShortName="unittest",
                          label=LangString(["unittest@en", "unittest@de"]),
                          namespaceIri=NamespaceIRI("http://unitest.org/project/unittest#"),
                          comment=LangString(["For testing@en", "Für Tests@de"]),
                          projectStart=Xsd_date(2024, 1, 1),
                          projectEnd=Xsd_date(2025, 12, 31)
                          )
        project.create()
        projectIri = project.projectIri
        self.assertIsNotNone(project.created)
        self.assertIsNotNone(project.creator)
        self.assertIsNotNone(project.modified)
        self.assertIsNotNone(project.contributor)
        del project

        project2 = Project.read(con=self._connection, projectIri_SName=projectIri, ignore_cache=True)
        self.assertEqual("unittest", project2.projectShortName)
        self.assertEqual(LangString(["unittest@en", "unittest@de"]), project2.label)
        self.assertEqual(LangString(["For testing@en", "Für Tests@de"]), project2.comment)
        self.assertEqual(Xsd_date(2024, 1, 1), project2.projectStart)
        self.assertEqual(Xsd_date(2025, 12, 31), project2.projectEnd)

        project3 = Project(con=self._connection,
                           projectShortName="unittest3",
                           label=LangString(["unittest3@en", "unittest3@de"]),
                           namespaceIri=NamespaceIRI("http://unitest.org/project/unittest3#"),
                           comment=LangString(["For testing3@en", "Für Tests3@de"]),
                           projectStart=Xsd_date(2024, 3, 3),
                           projectEnd=None)
        project3.create()
        projectIri3 = project3.projectIri
        project3 = Project.read(con=self._connection, projectIri_SName=projectIri3, ignore_cache=True)
        self.assertEqual("unittest3", project3.projectShortName)
        self.assertEqual(LangString(["unittest3@en", "unittest3@de"]), project3.label)
        self.assertEqual(LangString(["For testing3@en", "Für Tests3@de"]), project3.comment)
        self.assertEqual(Xsd_date(2024, 3, 3), project3.projectStart)

        project4 = Project(con=self._connection,
                           projectShortName="unittest4",
                           namespaceIri=NamespaceIRI("http://unitest.org/project/unittest4#"),
                           label=LangString("For testing4@en"))
        project4.create()
        projectIri4 = project4.projectIri
        project4 = Project.read(con=self._connection, projectIri_SName=projectIri4, ignore_cache=True)
        self.assertEqual("unittest4", project4.projectShortName)
        self.assertIsNone(project4.comment)
        self.assertIsNotNone(project4.projectStart)

        with self.assertRaises(OldapErrorInconsistency) as ex:
            project5 = Project(con=self._connection,
                               projectShortName="unittest5",
                               label=LangString(["unittest5@en", "unittest5@de"]),
                               namespaceIri=NamespaceIRI("http://unitest.org/project/unittest3#"),
                               comment=LangString(["For testing3@en", "Für Tests3@de"]),
                               projectStart=Xsd_date(2024, 3, 3),
                               projectEnd=Xsd_date(2024, 3, 2))

        project6 = Project(con=self._connection,
                           projectShortName="unittest6",
                           label=LangString("unittest\"; SELECT * WHERE {?s ?p ?o}"),
                           namespaceIri=NamespaceIRI("http://unitest.org/project/unittest6#"),
                           comment=LangString(["For testing6@en", "Für Tests6@de"]),
                           projectStart=Xsd_date(2024, 3, 3),
                           projectEnd=None)
        project6.create()
        projectIri = project6.projectIri
        del project6

        project6 = Project.read(con=self._connection, projectIri_SName=projectIri, ignore_cache=True)
        self.assertEqual(project6.label, LangString("unittest\"; SELECT * WHERE {?s ?p ?o}"))

        with self.assertRaises(OldapErrorInconsistency) as ex:
            project7 = Project(con=self._connection,
                               projectShortName="unittest7",
                               label=LangString("date"),
                               namespaceIri=NamespaceIRI("http://unitest.org/project/unittest3#"),
                               comment=LangString(["For testing3@en", "Für Tests3@de"]),
                               projectStart=Xsd_date(2024, 3, 3),
                               projectEnd=Xsd_date(2024, 3, 2))

        project8 = Project(con=self._connection,
                           projectShortName="unittest8",
                           label=LangString("unittes\"; SELECT * WHERE {?s ?p ?o}"),
                           namespaceIri=NamespaceIRI("http://unitest.org/project/unittest8#"),
                           comment=LangString(["For testing8@en", "Für Tests3@de"]),
                           projectEnd=Xsd_date(2028, 3, 3))
        self.assertEqual(project8.projectStart, date.today())

    def test_project_create_without_label_comment(self):
        project = Project(con=self._connection,
                          projectShortName="emptyfieldsXX",
                          namespaceIri=NamespaceIRI("http://unitest.org/project/unittest#"),
                          projectStart=Xsd_date(2024, 1, 1),
                          projectEnd=Xsd_date(2025, 12, 31)
                          )
        project.create()

    def test_project_create_empty_fields(self):
        project = Project(con=self._connection,
                          projectShortName="emptyfields1",
                          label=LangString(["unittest@en", "unittest@de"]),
                          namespaceIri=NamespaceIRI("http://unitest.org/project/unittest#"),
                          projectStart=Xsd_date(2024, 1, 1),
                          projectEnd=Xsd_date(2025, 12, 31)
                          )
        project.create()
        projectIri = project.projectIri
        del project
        project = Project.read(con=self._connection, projectIri_SName=projectIri, ignore_cache=True)
        self.assertIsNone(project.comment)

        project = Project(con=self._connection,
                          projectShortName="emptyfields2",
                          label=LangString(["unittest@en", "unittest@de"]),
                          namespaceIri=NamespaceIRI("http://unitest.org/project/unittest#"),
                          projectStart=Xsd_date(2024, 1, 1),
                          projectEnd=Xsd_date(2025, 12, 31)
                          )
        project.create()
        projectIri = project.projectIri
        del project
        project = Project.read(con=self._connection, projectIri_SName=projectIri, ignore_cache=True)
        project.comment = LangString("Comment for unittest@en", "Kommentar für unittest@de")
        self.assertEqual(project.comment[Language.EN], "Comment for unittest")
        self.assertEqual(project.comment[Language.DE], "Kommentar für unittest")
        project.update()
        del project
        project = Project.read(con=self._connection, projectIri_SName=projectIri, ignore_cache=True)
        self.assertEqual(project.comment[Language.EN], "Comment for unittest")
        self.assertEqual(project.comment[Language.DE], "Kommentar für unittest")

        project = Project(con=self._connection,
                          projectShortName="emptyfields3",
                          label=LangString(["unittest@en", "unittest@de"]),
                          comment=[],
                          namespaceIri=NamespaceIRI("http://unitest.org/project/unittest#"),
                          projectStart=Xsd_date(2024, 1, 1),
                          projectEnd=Xsd_date(2025, 12, 31)
                          )
        self.assertIsNone(project.comment)


    # @unittest.skip('Work in progress')
    def test_project_modify(self):
        project = Project(con=self._connection,
                          projectShortName="updatetest",
                          label=LangString(["updatetest@en", "updatetest@de"]),
                          namespaceIri=NamespaceIRI("http://unitest.org/project/updatetest#"),
                          comment=LangString(["For testing@en", "Für Tests@de"]),
                          projectStart=Xsd_date(2024, 1, 1),
                          projectEnd=Xsd_date(2025, 12, 31)
                          )
        project.create()

        projectIri = project.projectIri
        del project

        project = Project.read(con=self._connection, projectIri_SName=projectIri, ignore_cache=True)
        project.comment[Language.FR] = "Pour les tests"
        project.comment[Language.DE] = "FÜR DAS TESTEN"
        project.label = LangString(["UPDATETEST@en", "UP-DATE-TEST@fr"])
        project.projectEnd = Xsd_date(date(2026, 6, 30))
        project.update()
        self.assertEqual(project.comment, LangString(["For testing@en", "FÜR DAS TESTEN@de", "Pour les tests@fr"]))
        self.assertEqual(project.label, LangString(["UPDATETEST@en", "UP-DATE-TEST@fr"]))
        self.assertEqual(project.projectEnd, Xsd_date(2026, 6, 30))

    def test_project_update_B(self):
        project = Project(con=self._connection,
                          projectShortName="updatetestB",
                          label=LangString(["updatetest@en", "updatetest@de"]),
                          namespaceIri=NamespaceIRI("http://unitest.org/project/updatetestB#"),
                          comment=LangString(["For testing@en", "Für Tests@de"]),
                          projectStart=Xsd_date(2024, 1, 1),
                          projectEnd=Xsd_date(2025, 12, 31)
                          )
        project.create()

        projectIri = project.projectIri
        del project

        project = Project.read(con=self._connection, projectIri_SName="updatetestB", ignore_cache=True)
        project.comment[Language.IT] = "per i test"
        del project.comment[Language.EN]
        project.update()

        project = Project.read(con=self._connection, projectIri_SName="updatetestB", ignore_cache=True)
        self.assertEqual(project.comment, LangString(["Für Tests@de", "per i test@it"]))

    def test_project_start_end_consistency(self):
        project = Project(con=self._connection,
                          projectShortName="startendtest",
                          label=LangString(["startendtest@en", "startendtest@de"]),
                          namespaceIri=NamespaceIRI("http://unitest.org/project/startendtest#"),
                          comment=LangString(["For testing@en", "Für Tests@de"]),
                          projectStart=Xsd_date(2024, 1, 1),
                          projectEnd=Xsd_date(2025, 12, 31)
                          )
        with self.assertRaises(OldapErrorInconsistency):
            project.projectStart = Xsd_date(2026, 1, 1)
        with self.assertRaises(OldapErrorInconsistency):
            project.projectEnd = Xsd_date(2023, 12, 31)


    # @unittest.skip('Work in progress')
    def test_project_delete(self):
        project = Project(con=self._connection,
                          projectShortName="deletetest",
                          label=LangString(["deletetest@en", "deletetest@de"]),
                          namespaceIri=NamespaceIRI("http://unitest.org/project/deletetest#"),
                          comment=LangString(["For deleting@en", "Für Löschung@de"]),
                          projectStart=Xsd_date(2024, 1, 1),
                          #projectEnd=Xsd_date(2025, 12, 31)
                          )
        project.create()
        projectIri = project.projectIri

        project = Project.read(con=self._connection, projectIri_SName="deletetest", ignore_cache=True)
        project.delete()

        with self.assertRaises(OldapErrorNotFound) as ex:
            project = Project.read(con=self._connection, projectIri_SName=projectIri, ignore_cache=True)

    def test_unauthorized_access(self):
        project = Project(con=self._unpriv,
                          projectShortName="unauthorized",
                          label=LangString(["unauthorized@en", "unauthorized@de"]),
                          namespaceIri=NamespaceIRI("http://unitest.org/project/unauthorized#"),
                          comment=LangString(["For unauthorized access@en", "Für nicht authorisierten Zugang@de"]),
                          projectStart=Xsd_date(2024, 1, 1),
                          )
        cache = CacheSingleton()
        with self.assertRaises(OldapErrorNoPermission) as ex:
            project.create()

        project = Project(con=self._connection,
                          projectShortName="unauthorized",
                          label=LangString(["unauthorized@en", "unauthorized@de"]),
                          namespaceIri=NamespaceIRI("http://unitest.org/project/unauthorized#"),
                          comment=LangString(["For unauthorized access@en", "Für nicht authorisierten Zugang@de"]),
                          projectStart=Xsd_date(2024, 1, 1),
                          )
        project.create()

        project = Project.read(con=self._unpriv, projectIri_SName="unauthorized", ignore_cache=True)
        project.projectEnd = Xsd_date(2025, 12, 31)
        with self.assertRaises(OldapErrorNoPermission) as ex:
            project.update()
        with self.assertRaises(OldapErrorNoPermission) as ex:
            project.delete()


if __name__ == '__main__':
    unittest.main()



from typing import Optional, Self
from pystrict import strict

from oldaplib.src.enums.language import Language
from oldaplib.src.helpers.oldaperror import OldapErrorValue, OldapErrorIndex, OldapError
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd


#@strict
@serializer
class Xsd_string(Xsd):
    """
    Implements the XML Schema [xsd:string(https://www.w3.org/TR/xmlschema11-2/#string)] datatype.

    This class implements the handling of a single RDF/TRIG string that may have a language tag. It is
    serializable and manages the escaping/un-escaing of strings transfered to the triple store. This is
    an important part of the security conecpt of OLDAP.

    **No used-defined string may enter the triple store without being properly escaped to prevent
    SPARQL-injection!**

    The class implements the following methods:

    - `escaping(value: str)`: A static method that escapes a string
    - `unescaping(value: str)`: A static method that un-escapes a string
    - `Xsd_string(value: str, lang: Optional[str | Language] = None)`: Constructor with an optional
      language. The Language can be given either as [Language](/python_docstrings/language) enum or as
      2-letter ISO 639-1 language.
    - `fromRdf(value: str)`: A static constructor method that is used to convert an RDF/TRIG based string. It
      un-escapes the string retrieved from the triple store! This constructor is used by the
      [QueryProcessor](/python_docstrings/query_processor) class.
    - `str(Xsd_string)`: Returns a string with the langauge tag appended as `@ll`
      (ll as the 2-letter ISO languages), e.g. `'dies ist deutsch@de'`
    - `repr(Xsd_string)`: Returns a string with the langauge tag as it is beeiing used in TRIG/SPARQL
      queries, e.g. `'"dies ist deutsch"@de'`
    - `value`: Property returning the string value (without language tag)
    - `lang`: Property returning the language tag
    """
    __value: str | None
    __lang: Language | None

    @staticmethod
    def escaping(value: str) -> str:
        """
        Escape the given string to be suitable to be used in an RDF/SPARQL representation
        :param value: String ti be escaped
        :type value: str
        :return: Escaped string
        :rtype: str
        """
        value = value.replace('\\', '\\\\')
        value = value.replace("'", "\\'")
        value = value.replace('"', '\\"')
        value = value.replace('\n', '\\n').replace('\r', '\\r')
        return value

    @staticmethod
    def unescaping(value: str) -> str:
        """
        Unescape the given string
        :param value: Escaped string
        :type value: str
        :return: Unescpaed string
        :rtype: str
        """
        value = value.replace('\\\\', '\\')
        value = value.replace("\\'", "'")
        value = value.replace('\\"', '"')
        value = value.replace('\\n', '\n').replace('\\r', '\r')
        return value

    def __init__(self, value: Self | str | None = None, lang: str | Language | None = None, validate: bool = True):
        """
        Constructor of Xsd_string.
        :param value: String value
        :type value: str
        :param lang: Language the sting is in [optional]
        :type lang: str | Language
        :raises OldapErrorValue: If the given language is invalid
        """
        if not value:
            self.__value = None
            self.__lang = None
            return
        if isinstance(value, Xsd_string):
            self.__value = value.__value
            self.__lang = value.__lang
        else:
            value = str(value)
            if lang is None:
                if len(value) > 3 and value[-3] == '@':
                    tmpls: str = value[-2:].upper()
                    try:
                        self.__lang = Language[tmpls]
                        self.__value = value[:-3]
                    except KeyError as err:
                        self.__lang = None
                        self.__value = value
                else:
                    self.__lang = None
                    self.__value = value
                return
            self.__value = value
            if isinstance(lang, Language):
                self.__lang = lang
            else:
                try:
                    self.__lang = Language[lang.upper()] if lang else None
                except (ValueError, KeyError) as err:
                    raise OldapErrorValue(str(err))

    @classmethod
    def fromRdf(cls, value: str, lang: Optional[str] = None) -> Self:
        """
        Constructor of Xsd_string that takes a RDF/SPARQL representation of the string and un-escapes it
        :param value: String from RDF/SPARQL representation
        :type value: str
        :param lang: Language ISO short (lowercased) [optional]
        :type lang: str
        :return: OldapStringLiteral instance
        :rtype: StringLiteral
        """
        value = Xsd_string.unescaping(value)
        return cls(value, lang, validate=False)

    def __len__(self) -> int:
        if self.__value is None:
            return 0
        return len(self.__value)

    def __bool__(self) -> bool:
        return self.__value is not None

    def __str__(self) -> str:
        """
        Returns the string representation of the Xsd_string instance with the languages appended "@ll"
        :return: Language string
        :rtype: str
        """
        if self.__value is None:
            return 'None'
        if self.__lang:
            return f'{self.__value}@{self.__lang.name.lower()}'
        else:
            return self.__value

    def __repr__(self) -> str:
        """
        Returns the Xsd_string instance as string as used in a RDF/SPARQL representation
        :return: SPARQL/RDF/TRIG representation of the language string
        :rtype: str
        """
        if self.__value is None:
            return 'None'
        if self.__lang:
            return f'Xsd_string("{self.__value}", "{self.__lang.name.lower()}")'
        else:
            return f'Xsd_string("{self.__value}")'

    def __eq__(self, other: str | Self | None) -> bool:
        """
        Check for equality of two Xsd_strings (both strings and languages must be equal)
        :param other: Other OldapStringLiteral
        :type other: StringLiteral
        :return: True or False
        :rtype: bool
        """
        if other is None:
            if self.__value is None:
                return True
            else:
                return False
        if isinstance(other, Xsd_string):
            return self.__value == other.__value and self.__lang == other.__lang
        elif isinstance(other, str):
            return self.__value == str(other)

    def __ne__(self, other: str | Self | None) -> bool:
        """
        Check for inequality of two Xsd_strings (both strings and languages must be equal)
        :param other: An Xsd_string or a string value
        :type other: Xsd_string
        :return: True or False
        :rtype: bool
        """
        if other is None:
            if self.__value is None:
                return False
            else:
                return True
        if isinstance(other, Xsd_string):
            return self.__value != other.__value and self.__lang == other.__lang
        elif isinstance(other, str):
            return self.__value != other

    def __getitem__(self, key: int | slice) -> str | None:
        """
        Get the character(s) at the given position (slicing or indexing supported).
        :param key: Index or slice
        :type key: int | slice
        :return: Character or slice as string
        :rtype: str | None
        """
        c = None
        try:
            c = self.__value[key]
        except (IndexError, TypeError) as err:
            raise OldapErrorIndex(f'Cannot get character from Xsd_string at possition {key}: {err}')
        return c

    def __hash__(self) -> int:
        """
        Returns the Xsd_string hash value as an integer
        :return: Hash value
        :rtype: int
        """
        if self.__lang:
            return hash(self.__value + '@' + self.__lang.value)
        else:
            return hash(self.__value)

    @property
    def toRdf(self):
        """
        RDF representation of the Xsd_string instance
        :return: RDF representation of the Xsd_string instance
        :rtype: str
        """
        if self.__value is None:
            raise OldapErrorValue(f'Cannot convert empty Xsd_string to RDF string.')
        if self.__lang:
            return f'"{Xsd_string.escaping(self.__value)}"@{self.__lang.name.lower()}'
        else:
            return f'"{Xsd_string.escaping(self.__value)}"^^xsd:string'

    def _as_dict(self) -> dict:
        """
        Used for the JSON serializer (@serializer decorator)
        :return: Dictionary representation of the OldapStringLiteral
        :rtype: dict
        """
        return {
                "value": self.__value,
                "lang": self.__lang
        }

    @property
    def value(self) -> str | None:
        """
        Returns the Xsd_string string value only
        :return: string representation of the Xsd_string without language
        :rtype: str
        """
        return self.__value

    @property
    def lang(self) -> Language | None:
        """
        Returns the Language if available (or None
        :return: Language
        :rtype: Language | None
        """
        return self.__lang



import unittest
from pathlib import Path
from time import sleep

from oldaplib.src.connection import Connection
from oldaplib.src.dtypes.namespaceiri import NamespaceIRI
from oldaplib.src.enums.language import Language
from oldaplib.src.helpers.context import Context
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.helpers.oldaperror import OldapErrorNotFound, OldapErrorImmutable, OldapErrorNoPermission
from oldaplib.src.iconnection import IConnection
from oldaplib.src.oldaplist import OldapList
from oldaplib.src.enums.oldaplistattr import OldapListAttr
from oldaplib.src.project import Project
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_string import Xsd_string


def find_project_root(current_path):
    # Climb up the directory hierarchy and check for a marker file
    path = Path(current_path).absolute()
    while not (path / 'pyproject.toml').exists():
        if path.parent == path:
            # Root of the filesystem, file not found
            raise RuntimeError('Project root not found')
        path = path.parent
    return path


class TestOldapList(unittest.TestCase):

    _connection: IConnection
    _unpriv: Connection

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        project_root = find_project_root(__file__)

        cls._context = Context(name="DEFAULT")
        cls._context['test'] = NamespaceIRI("http://testing.org/datatypes#")
        cls._context.use('test')
        cls._connection = Connection(server='http://localhost:7200',
                                     repo="oldap",
                                     userId="rosenth",
                                     credentials="RioGrande",
                                     context_name="DEFAULT")
        cls._unpriv = Connection(server='http://localhost:7200',
                                 repo="oldap",
                                 userId="fornaro",
                                 credentials="RioGrande",
                                 context_name="DEFAULT")

        cls._connection.clear_graph(Xsd_QName('oldap:admin'))
        file = project_root / 'oldaplib' / 'ontologies' / 'admin.trig'
        cls._connection.upload_turtle(file)


        cls._connection.clear_graph(Xsd_QName('test:test'))
        cls._connection.clear_graph(Xsd_QName('test:onto'))
        cls._connection.clear_graph(Xsd_QName('test:shacl'))
        cls._connection.clear_graph(Xsd_QName('test:lists'))
        file = project_root / 'oldaplib' / 'testdata' / 'connection_test.trig'
        cls._connection.upload_turtle(file)
        sleep(1)
        cls._project = Project.read(cls._connection, "test")
        LangString.defaultLanguage = Language.EN


    @classmethod
    def tearDownClass(cls):
        pass

    def test_constructor_project_shortname(self):
        oldaplist = OldapList(con=self._connection,
                              project="test",
                              oldapListId="TestList",
                              prefLabel="TestList",
                              definition="A list for testing...")
        self.assertEqual(Xsd_NCName('TestList'), oldaplist.oldapListId)
        self.assertEqual(LangString("TestList"), oldaplist.prefLabel)
        self.assertEqual(LangString("A list for testing..."), oldaplist.definition)
        self.assertEqual(NamespaceIRI("http://oldap.org/test/TestList#"), oldaplist.node_namespaceIri)

    def test_constructor_project_object(self):
        oldaplist = OldapList(con=self._connection,
                              project=self._project,
                              oldapListId="TestList2",
                              prefLabel="TestList2",
                              definition="A list for testing...")
        self.assertEqual(Xsd_NCName('TestList2'), oldaplist.oldapListId)
        self.assertEqual(LangString("TestList2"), oldaplist.prefLabel)
        self.assertEqual(LangString("A list for testing..."), oldaplist.definition)
        self.assertEqual(NamespaceIRI("http://oldap.org/test/TestList2#"), oldaplist.node_namespaceIri)

    def test_create_read_project_id(self):
        oldaplist = OldapList(con=self._connection,
                              project="test",
                              oldapListId="TestList_B",
                              prefLabel="TestList_B",
                              definition="A list for testing...")
        oldaplist.create()
        del oldaplist
        oldaplist = OldapList.read(con=self._connection,
                                   project="test",
                                   oldapListId="TestList_B")
        self.assertEqual("TestList_B", oldaplist.oldapListId)
        self.assertEqual(LangString("TestList_B"), oldaplist.prefLabel)
        self.assertEqual(LangString("A list for testing..."), oldaplist.definition)
        self.assertEqual(NamespaceIRI("http://oldap.org/test/TestList_B#"), oldaplist.node_namespaceIri)

    def test_create_unpriv(self):
        oldaplist = OldapList(con=self._unpriv,
                              project="test",
                              oldapListId="TestList_Unpriv",
                              prefLabel="TestList_Unpriv",
                              definition="A list for testing...")
        with self.assertRaises(OldapErrorNoPermission) as ex:
            oldaplist.create()

    def test_create_read_project_object(self):
        oldaplist = OldapList(con=self._connection,
                              project=self._project,
                              oldapListId="TestList_A",
                              prefLabel="TestList_A",
                              definition="A list for testing...")
        oldaplist.create()
        del oldaplist
        oldaplist = OldapList.read(con=self._connection,
                                   project=self._project,
                                   oldapListId="TestList_A")
        self.assertEqual("TestList_A", oldaplist.oldapListId)
        self.assertEqual(LangString("TestList_A"), oldaplist.prefLabel)
        self.assertEqual(LangString("A list for testing..."), oldaplist.definition)
        self.assertEqual(NamespaceIRI("http://oldap.org/test/TestList_A#"), oldaplist.node_namespaceIri)

    def test_update_A(self):
        oldaplist = OldapList(con=self._connection,
                              project=self._project,
                              oldapListId="TestUpdateListA",
                              prefLabel="TestUpdateListA",
                              definition="A list for testing updates...")
        oldaplist.create()
        del oldaplist
        oldaplist = OldapList.read(con=self._connection,
                                   project=self._project,
                                   oldapListId="TestUpdateListA")
        self.assertEqual("TestUpdateListA", oldaplist.oldapListId)
        self.assertEqual(LangString("TestUpdateListA"), oldaplist.prefLabel)
        self.assertEqual(LangString("A list for testing updates..."), oldaplist.definition)

        oldaplist.prefLabel[Language.FR] = "TestesDeModifications"
        oldaplist.definition = LangString("Test-A@en", "test-B@fr", "test-C@de")
        oldaplist.update()
        oldaplist = OldapList.read(con=self._connection,
                                   project=self._project,
                                   oldapListId="TestUpdateListA")
        self.assertEqual(oldaplist.prefLabel, LangString("TestUpdateListA@en", "TestesDeModifications@fr"))
        self.assertEqual(oldaplist.definition, LangString("Test-A@en", "test-B@fr", "test-C@de"))

        with self.assertRaises(OldapErrorImmutable):
            oldaplist.oldapListId = "Gagagag"

    def test_update_B(self):
        oldaplist = OldapList(con=self._connection,
                              project="test",
                              oldapListId="TestUpdateListB",
                              prefLabel="TestUpdateListB")
        oldaplist.create()
        del oldaplist
        oldaplist = OldapList.read(con=self._connection,
                                   project=self._project,
                                   oldapListId="TestUpdateListB")
        oldaplist.definition=LangString("Test-A@en", "test-B@fr", "test-C@de")
        oldaplist.update()
        oldaplist = OldapList.read(con=self._connection,
                                   project=self._project,
                                   oldapListId="TestUpdateListB")
        self.assertEqual(oldaplist.definition, LangString("Test-A@en", "test-B@fr", "test-C@de"))

    def test_update_C(self):
        oldaplist = OldapList(con=self._connection,
                              project=self._project,
                              oldapListId="TestUpdateListC",
                              prefLabel="TestUpdateListC",
                              definition="A list for testing updates...")
        oldaplist.create()
        del oldaplist
        oldaplist = OldapList.read(con=self._connection,
                                   project=self._project,
                                   oldapListId="TestUpdateListC")
        del oldaplist.definition
        oldaplist.update()
        oldaplist = OldapList.read(con=self._connection,
                                   project=self._project,
                                   oldapListId="TestUpdateListC")
        self.assertFalse(oldaplist.get(OldapListAttr.DEFINITION))

    def test_delete(self):
        oldaplist = OldapList(con=self._connection,
                              project=self._project,
                              oldapListId="TestDeleteList",
                              prefLabel="TestDeleteList",
                              definition="A list for testing deletes...")
        oldaplist.create()
        del oldaplist
        oldaplist = OldapList.read(con=self._connection,
                                   project=self._project,
                                   oldapListId="TestDeleteList")
        oldaplist.delete()
        with self.assertRaises(OldapErrorNotFound) as ex:
            oldaplist = OldapList.read(con=self._connection,
                                       project=self._project,
                                       oldapListId="TestDeleteList")

    def test_search(self):
        oldaplist = OldapList(con=self._connection,
                              project=self._project,
                              oldapListId="animals",
                              prefLabel=LangString("Animals@en", "Tiere@de", "Animaux@fr"),
                              definition="A hierarchical list of all animals")
        oldaplist.create()
        oldaplist = OldapList(con=self._connection,
                              project=self._project,
                              oldapListId="plants",
                              prefLabel=LangString("Plants@en", "Pflanzen@de"),
                              definition="A hierarchical list of all plants")
        oldaplist.create()

        iris = OldapList.search(con=self._connection,
                                project=self._project,
                                prefLabel=Xsd_string("Animals", "en"))
        self.assertEqual(iris, [Iri('test:animals')])

        iris = OldapList.search(con=self._connection,
                                project=self._project,
                                prefLabel=Xsd_string("Animals"))
        self.assertEqual(iris, [Iri('test:animals')])

        iris = OldapList.search(con=self._connection,
                                project=self._project,
                                prefLabel=Xsd_string("Pflanzen", "de"))
        self.assertEqual(iris, [Iri('test:plants')])

        iris = OldapList.search(con=self._connection,
                                project=self._project,
                                prefLabel=Xsd_string("Tier"))
        self.assertEqual(iris, [Iri('test:animals')])

        iris = OldapList.search(con=self._connection,
                                project=self._project,
                                prefLabel=Xsd_string("Tier"),
                                exactMatch=True)
        self.assertEqual(iris, [])

        iris = OldapList.search(con=self._connection,
                                project=self._project,
                                prefLabel=Xsd_string("Pilze"))
        self.assertEqual(iris, [])

        iris = OldapList.search(con=self._connection,
                                project=self._project,
                                definition=Xsd_string("hierarchical"))
        self.assertTrue(Iri('test:animals') in iris)
        self.assertTrue(Iri('test:plants') in iris)




if __name__ == '__main__':
    unittest.main()
